import enum


@enum.unique
class Status(enum.Enum):
    """
    Status of a Recipe denoting whether (re)evaluation is needed
    """
    Ok = 0  # Recipe does not need (re)evaluation

    # The following all require (re)evaluation
    IngredientDirty = 1  # One or more ingredients of the recipe have changed
    NotEvaluatedYet = 2  # Recipe has not been evaluated yet
    InputsChanged = 3  # One or more inputs to the recipe have changed
    OutputsInvalid = 4  # One or more outputs of the recipe have been changed externally
    BoundFunctionChanged = 5  # The function referenced by the recipe has changed
    CustomDirty = 6  # The recipe has been marked dirty through a custom cleanliness function



import argparse
import logging
import sys
from typing import Dict, Union, Any, List, Iterable, Optional, TextIO

from .core import Status, compute_recipe_status, create_graph
from .logging import log
from .recipe import Recipe
from .recipes import Arg


class Lab:
    """
    Class used to define a collection of alkymi recipes and expose them as a command line interface (CLI)

    This can be used to create files that bear resemblance to Makefiles (see alkymi/labfile.py as an example)
    """

    def __init__(self, name: str):
        """
        Creates a new Lab with the provided name

        :param name: The name of the Lab
        """
        self._name = name
        self._recipes: List[Recipe] = []
        self._args: Dict[str, Arg] = {}

    def add_recipe(self, recipe: Recipe) -> Recipe:
        """
        Add a new recipe to the Lab (this will make the recipe available through the CLI)

        :param recipe: The recipe to add
        :return: The input recipe (to allow chaining calls)
        """
        if recipe not in self._recipes:
            self._recipes.append(recipe)
        return recipe

    def add_recipes(self, *recipes: Recipe) -> None:
        """
        Add a set of recipes to the Lab (this will make the recipes available through the CLI)

        :param recipes: The recipes to add
        """
        for recipe in recipes:
            self.add_recipe(recipe)

    def register_arg(self, arg: Arg) -> None:
        """
        Register an argument with the Lab (this will make the argument settable through the CLI)

        :param arg: The argument to register
        """
        self._args[arg.name] = arg

    def brew(self, target_recipe: Union[Recipe, str], *, jobs=1) -> Any:
        """
        Brew (evaluate) a target recipe defined by its reference or name, and return the results

        :param target_recipe: The recipe to evaluate, as a reference ot by name
        :param jobs: The number of jobs to use for evaluating this recipe in parallel, defaults to 1 (no parallelism),
                     zero or negative values will cause alkymi to use the system's default number of jobs
        :return: The output of the evaluated recipe
        """
        if isinstance(target_recipe, str):
            # Try to match name
            for recipe in self._recipes:
                if recipe.name == target_recipe:
                    return recipe.brew(jobs=jobs)
            raise ValueError("Unknown recipe: {}".format(target_recipe))
        else:
            # Match recipe directly
            if target_recipe in self._recipes:
                return target_recipe.brew(jobs=jobs)
            raise ValueError("Unknown recipe: {}".format(target_recipe.name))

    @property
    def name(self) -> str:
        """
        :return: The name of this Lab
        """
        return self._name

    @property
    def recipes(self) -> List[Recipe]:
        """
        :return: The list of recipes contained in this Lab
        """
        return self._recipes

    @property
    def args(self) -> Dict[str, Arg]:
        """
        :return: The list of args registered with this Lab
        """
        return self._args

    def _build_full_status(self) -> Dict[Recipe, Status]:
        """
        Compute statuses for all recipes (and dependent recipes) in this Lab

        :return: The statuses as a dictionary
        """
        status: Dict[Recipe, Status] = {}
        for recipe in self._recipes:
            graph = create_graph(recipe)
            status.update(compute_recipe_status(recipe, graph))
        return status

    def _add_user_args_(self, parser: argparse.ArgumentParser) -> None:
        """
        Adds user provided arguments to an ArgumentParser instance

        :param parser: The parser to add the user-provided arguments to
        """
        for arg_name, arg in self._args.items():
            # For iterables (e.g. lists), the "type" keyword is actually the type of elements in the iterable
            if issubclass(arg.type, Iterable) and not arg.type == str:
                subtype = arg.subtype if arg.subtype is not None else str
                parser.add_argument("--{}".format(arg_name), type=subtype, nargs="*", dest=arg_name)
            else:
                parser.add_argument("--{}".format(arg_name), type=arg.type, dest=arg_name)

    def __repr__(self) -> str:
        """
        :return: A string representation of this Lab with recipes and their statuses
        """
        status = self._build_full_status()
        state = ''
        for recipe in self._recipes:
            state += '\n\t{} - {}'.format(recipe.name, status[recipe])
        return '{} lab with recipes:{}'.format(self.name, state)

    def open(self, args: Optional[List[str]] = None, stream: TextIO = sys.stderr) -> None:
        """
        Runs the command line interface for this Lab by parsing command line arguments and carrying out the designated
        command

        :param args: The input arguments to use - will default to system args
        :param stream: The stream to print output to
        """
        if len(self.recipes) == 0:
            raise RuntimeError("No recipes added to lab - CLI is useless")

        # Use system args if nothing has been provided
        if args is None:
            args = sys.argv[1:]

        # Create the top-level parser
        parser = argparse.ArgumentParser('CLI for {}'.format(self._name))
        parser.add_argument("-v", "--verbose", action="store_true", help="Turn on verbose logging")

        subparsers = parser.add_subparsers(help='sub-command help', dest='subparser_name')

        # Create the parser for the "status" command
        status_parser = subparsers.add_parser('status', help='Prints the detailed status of the lab')
        self._add_user_args_(status_parser)

        # Create the parser for the "brew" command
        brew_parser = subparsers.add_parser('brew', help='Brew the selected recipe')
        brew_parser.add_argument('recipe', choices=[recipe.name for recipe in self._recipes], nargs="+",
                                 help='Recipe(s) to brew')
        brew_parser.add_argument("-j", "--jobs", type=int, default=1,
                                 help="Use N jobs to evaluate the recipe, more than 1 job will parallelize evaluation")
        self._add_user_args_(brew_parser)

        parsed_args = parser.parse_args(args)
        log.addHandler(logging.StreamHandler(stream))
        if parsed_args.verbose:
            log.setLevel(logging.DEBUG)
        else:
            log.setLevel(logging.INFO)

        # Set arguments if supplied
        for arg_name, arg in self._args.items():
            provided_val = getattr(parsed_args, arg_name, None)
            if provided_val is not None:
                arg.set(provided_val)

        if parsed_args.subparser_name == 'status':
            print(self, file=stream)
        elif parsed_args.subparser_name == 'brew':
            for recipe in parsed_args.recipe:
                self.brew(recipe, jobs=parsed_args.jobs)
        else:
            # No recognized command provided - print help
            parser.print_help(file=stream)



import json
from collections import OrderedDict
from pathlib import Path
from typing import Iterable, Callable, List, Optional, Tuple, TypeVar, Generic, cast

from . import checksums, serialization
from .config import CacheType, AlkymiConfig
from .logging import log
from .serialization import OutputWithValue, CachedOutput, Output
from .types import Status

R = TypeVar("R")  # The return type of the bound function

CleanlinessFunc = Callable[[R], bool]


class Recipe(Generic[R]):
    """
    Recipe is the basic building block of alkymi's evaluation approach. It binds a function (provided by the user) that
    it then calls when asked to by alkymi's execution engine. The result of the bound function evaluation can be
    automatically cached to disk to allow for checking of cleanliness (whether a Recipe is up-to-date), and to avoid
    invoking the bound function if necessary on subsequent evaluations
    """

    CACHE_DIRECTORY_NAME = ".alkymi_cache"

    def __init__(self, func: Callable[..., R], ingredients: Iterable['Recipe'], name: str, transient: bool,
                 cache: CacheType, cleanliness_func: Optional[CleanlinessFunc[R]] = None):
        """
        Create a new Recipe

        :param ingredients: The dependencies of this Recipe - the outputs of these Recipes will be provided as arguments
                            to the bound function when called (following the item from the mapped_inputs sequence)
        :param func: The function to bind to this recipe
        :param name: The name of this Recipe
        :param transient: Whether to always (re)evaluate the created Recipe
        :param cache: The type of caching to use for this Recipe
        :param cleanliness_func: A function to allow a custom cleanliness check
        """
        self._func = func
        self._ingredients = list(ingredients)
        self._name = name
        self._transient = transient
        self._cleanliness_func = cleanliness_func

        # Set cache type based on default value (in AlkymiConfig)
        if cache == CacheType.Auto:
            # Pick based on what is in the config
            self._cache = CacheType.Cache if AlkymiConfig.get().cache else CacheType.NoCache
        else:
            self._cache = cache

        self._outputs: Optional[Output[R]] = None
        self._input_checksums: Optional[Tuple[Optional[str], ...]] = None
        self._last_function_hash: Optional[str] = None

        if self.cache == CacheType.Cache:
            # Try to reload last state
            func_file = Path(self._func.__code__.co_filename)
            module_name = func_file.absolute().parent.stem

            # Use the cache path set in the alkymi config, or fall back to current working dir
            cache_root = AlkymiConfig.get().cache_path
            if cache_root is None:
                cache_root = Path(".")
            self.cache_path = (cache_root / Recipe.CACHE_DIRECTORY_NAME / module_name / name).absolute()

            self.cache_file = self.cache_path / 'cache.json'
            if self.cache_file.exists():
                with self.cache_file.open('r') as f:
                    self.restore_from_dict(json.load(f))

    def __call__(self, *args) -> R:
        """
        Calls the bound function directly

        :param args: The arguments to provide to the bound function
        :param kwargs: The keyword arguments to provide to the bound function
        :return: The value returned by the bound function
        """
        return self._func(*args)

    def set_result(self, outputs: R, input_checksums: Tuple[Optional[str], ...]) -> None:
        """
        Stores the provided result in the recipe and caches it to disk if applicable

        :param outputs: The outputs to store in the recipe
        :param input_checksums: The checksums of the inputs that were used to calculate the outputs
        """
        self.outputs = outputs
        self._input_checksums = input_checksums
        self._last_function_hash = self.function_hash
        self._save_state()

    def brew(self, *, jobs: int = 1) -> R:
        """
        Evaluate this Recipe and all dependent inputs - this will build the computational graph and execute any needed
        dependencies to produce the outputs of this Recipe

        :param jobs: The number of jobs to use for evaluating this recipe in parallel, defaults to 1 (no parallelism),
                     zero or negative values will cause alkymi to use the system's default number of jobs
        :return: The outputs of this Recipe (which correspond to the outputs of the bound function)
        """
        # Lazy import to avoid circular imports
        from .core import brew
        return brew(self, jobs=jobs)

    def status(self) -> Status:
        """
        :return: The status of this recipe (will evaluate all upstream dependencies)
        """
        # Lazy import to avoid circular imports
        from .core import compute_recipe_status, create_graph
        return compute_recipe_status(self, create_graph(self))[self]

    def __str__(self) -> str:
        return self.name

    def _save_state(self) -> None:
        """
        Save the current state of this Recipe to a json file and zero or more extra data files (as needed)
        """
        if self._cache == CacheType.Cache:
            self.cache_path.mkdir(exist_ok=True, parents=True)
            with self.cache_file.open('w') as f:
                f.write(json.dumps(self.to_dict(), indent=4))

    @property
    def outputs_valid(self) -> bool:
        """
        Check whether an output is still valid - this is currently only used to check files that may have been deleted
        or altered outside alkymi's cache. If no outputs have been produced yet, True will be returned.

        :return: Whether all outputs are still valid
        """
        if self._outputs is None:
            return True
        return self._outputs.valid

    @property
    def custom_cleanliness_func(self) -> Optional[CleanlinessFunc]:
        return self._cleanliness_func

    @property
    def last_function_hash(self) -> Optional[str]:
        return self._last_function_hash

    @property
    def name(self) -> str:
        """
        :return: The name of this Recipe
        """
        return self._name

    @property
    def ingredients(self) -> List['Recipe']:
        """
        :return: The dependencies of this Recipe - the outputs of these Recipes will be provided as arguments to the
                 bound function when called (following the item from the mapped_inputs sequence)
        """
        return self._ingredients

    @property
    def transient(self) -> bool:
        """
        :return: Whether to always (re)evaluate the created Recipe
        """
        return self._transient

    @property
    def cache(self) -> CacheType:
        """
        :return: The type of caching to use for this Recipe
        """
        return self._cache

    @property
    def function_hash(self) -> str:
        """
        :return: The hash of the bound function as a string
        """
        return checksums.function_hash(self._func)

    @property
    def input_checksums(self) -> Optional[Tuple[Optional[str], ...]]:
        """
        :return: The checksum(s) for the inputs
        """
        return self._input_checksums

    @property
    def outputs(self) -> Optional[R]:
        """
        :return: The outputs of this Recipe
        """
        if self._outputs is None:
            return None
        return self._outputs.value()

    @outputs.setter
    def outputs(self, outputs: R) -> None:
        """
        Sets the outputs of this Recipe and computes the necessary checksums needed for checking dirtiness

        :param outputs: outputs of this Recipe
        """
        self._outputs = OutputWithValue(outputs, checksums.checksum(outputs))

    @property
    def output_checksum(self) -> Optional[str]:
        """
        :return: The computed checksums for the outputs (this is set when outputs are set)
        """
        if self._outputs is None:
            return None
        return self._outputs.checksum

    def to_dict(self) -> OrderedDict:
        """
        :return: The Recipe as a dict for serialization purposes
        """
        # Force caching of all outputs (if they aren't already)
        serialized_outputs = None
        if self._outputs is not None:
            if isinstance(self._outputs, CachedOutput):
                pass
            elif isinstance(self._outputs, OutputWithValue):
                self._outputs = serialization.cache(self._outputs, self.cache_path)
            else:
                raise RuntimeError("Output is of wrong type")
            serialized_outputs = self._outputs.serialized

        return OrderedDict(
            name=self.name,
            input_checksums=self.input_checksums,
            outputs=serialized_outputs,
            output_checksum=self.output_checksum,
            last_function_hash=self._last_function_hash,
        )

    def restore_from_dict(self, old_state) -> None:
        """
        Restores the state of this Recipe from a previously cached state

        :param old_state: The old cached state to restore
        """
        log.debug("Restoring {} from dict".format(self._name))
        if old_state["input_checksums"] is not None:
            self._input_checksums = tuple(old_state["input_checksums"])
        if old_state["outputs"] is not None and old_state["output_checksum"] is not None:
            self._outputs = CachedOutput(None, old_state["output_checksum"], old_state["outputs"])
        self._last_function_hash = cast(str, old_state["last_function_hash"])



from pathlib import Path
from typing import Any, Sequence, Dict, Callable
import pickle
import alkymi.config

# Try to use xxh3 from xxhash to speed up hashing significantly, otherwise fallback to built-in MD5
try:
    import xxhash

    HASHER = xxhash.xxh3_64
except ImportError:
    import hashlib

    HASHER = hashlib.md5

import inspect

# Load additional checksum generators based on available libs
from alkymi import AlkymiConfig

additional_checksum_generators: Dict[Any, Callable] = {}
try:
    import numpy as np  # NOQA


    def _handle_ndarray(array: np.ndarray) -> str:
        """
        Computes a checksum for the provided numpy array

        :param array: The numpy array to compute a checksum for
        :return: The computed checksum as a string
        """
        return HASHER(array.data).hexdigest()


    additional_checksum_generators[np.ndarray] = _handle_ndarray
except ImportError:
    pass

try:
    import pandas as pd  # NOQA


    def _handle_dataframe(df: pd.DataFrame) -> str:
        """
        Computes a checksum for the provided pandas DataFrame

        :param df: The pandas DataFrame to compute a checksum for
        :return: The computed checksum as a string
        """
        return HASHER(pd.util.hash_pandas_object(df).values).hexdigest()


    additional_checksum_generators[pd.DataFrame] = _handle_dataframe
except ImportError:
    pass


class Checksummer(object):
    """
    Class used to compute a stable hash/checksum of an object recursively. Currently uses MD5.
    """

    def __init__(self):
        self._hasher = HASHER()

    def update(self, obj: Any) -> None:
        """
        Update the current checksum with new information from the provided value. May call itself recursively if needed
        to hash complex inputs.

        :param obj: The object to update the checksum with
        """
        if obj is None:
            return

        # The type of the input object needs to be taken into consideration to avoid different types with the same value
        # resulting in the same checksum
        self._hasher.update(str(type(obj)).encode("utf-8"))

        if isinstance(obj, str):
            self._hasher.update(obj.encode("utf-8"))
        elif isinstance(obj, bytes):
            self._hasher.update(obj)
        elif isinstance(obj, (int, float)):
            self.update(str(obj))
        elif isinstance(obj, Sequence):
            for e in obj:
                self.update(e)
        elif isinstance(obj, Dict):
            keys = obj.keys()
            for k in keys:
                self.update(k)
                self.update(obj[k])
        elif isinstance(obj, Path):
            self._update_path(obj)
        elif inspect.iscode(obj):
            self.update(obj.co_code)
        elif inspect.isroutine(obj):
            self._update_func(obj)
        else:
            # Check if any additional checksum generator will work
            generator = additional_checksum_generators.get(type(obj))
            if generator is not None:
                self.update(generator(obj))
            else:
                # As a last resort, try to pickle the object to get bytes for hashing
                if not AlkymiConfig.get().allow_pickling:
                    raise RuntimeError("Pickling disabled - cannot checksum item: {}".format(type(obj)))

                try:
                    pickled_bytes = pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)
                    self.update(pickled_bytes)
                except pickle.PicklingError:
                    raise ValueError("Checksum not supported for type: {}".format(type(obj)))

    def _update_func(self, fn) -> None:
        """
        Update the current checksum with a function

        :param fn: The function to update the checksum with
        """
        code = fn.__code__

        # Hash the bytecode
        self.update(code.co_code)

        # Hash default arguments
        defaults = fn.__defaults__
        if defaults is not None:
            default_values = dict(zip(code.co_varnames[-len(defaults):], defaults))
            self.update(default_values)

        # Hash constants that are referenced by the bytecode but ignore names of lambdas
        if code.co_consts:
            for const in code.co_consts:
                if not isinstance(const, str) or not const.endswith(".<lambda>"):
                    self.update(const)

        # Handle referenced functions
        if code.co_freevars:
            assert len(code.co_freevars) == len(fn.__closure__)
            referenced_func_names_and_closures = list(
                zip(code.co_freevars, (c.cell_contents for c in fn.__closure__)))
            self.update(referenced_func_names_and_closures)

    def _update_path(self, path: Path) -> None:
        """
        Update the current checksum with a Path

        :param path: The Path to update the checksum with
        """
        # For non-existent paths, we just care about the path itself
        if not path.exists():
            self.update(str(path))
            return

        # For directories, we care about the path and whether it exists
        if path.is_dir():
            self.update(str(path))
            self.update(path.exists())
            return

        # For files, we care about file name ...
        self.update(str(path))

        # ... and either the file hash
        file_checksum_method = alkymi.config.AlkymiConfig.get().file_checksum_method
        if file_checksum_method == alkymi.config.FileChecksumMethod.HashContents:
            with path.open('rb') as f:
                size = 1024 * self._hasher.block_size
                b = f.read(size)
                while len(b) > 0:
                    self._hasher.update(b)
                    b = f.read(size)

        # ... or the file modification timestamp
        elif file_checksum_method == alkymi.config.FileChecksumMethod.ModificationTimestamp:
            last_modification_timestamp = path.stat().st_mtime_ns
            self.update(last_modification_timestamp)

    def digest(self) -> str:
        """
        :return: The checksum as a string
        """
        return self._hasher.hexdigest()


def function_hash(fn: Callable) -> str:
    """
    Computes the hash/checksum of a function

    :param fn: The function to hash
    :return: The checksum as a string
    """
    hasher = Checksummer()
    hasher.update(fn)
    return hasher.digest()


def checksum(obj: Any) -> str:
    """
    Computes the hash/checksum of the provided input

    :param obj: The object to compute a hash/checksum for
    :return: The checksum as a string
    """
    # Fake the hash of None as just "None"
    if obj is None:
        return "None"

    hasher = Checksummer()
    hasher.update(obj)
    return hasher.digest()



import subprocess
import sys
from typing import List, TextIO, Optional

from .logging import log


def call(args: List[str], echo_error_to_stream: Optional[TextIO] = sys.stderr,
         echo_output_to_stream: Optional[TextIO] = sys.stdout) -> subprocess.CompletedProcess:
    """
    Utility command to run a system command and return the result (stdout and stderr will also be printed to the debug
    log)

    :param args: The arguments representing the command to run, e.g. ["echo", "test"]
    :param echo_error_to_stream: A stream to which to echo the call's stderr on a non-zero exit code
    :param echo_output_to_stream: A stream to which to echo the call's stdout while the command is executing
    :return: The result of the execution as a subprocess.CompletedProcess instance
    """
    # Buffer one line at a time if echoing live, otherwise just use the default
    buffer_size = 1 if echo_output_to_stream is not None else -1
    proc = subprocess.Popen(args, universal_newlines=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                            bufsize=buffer_size)

    # Since we set these to PIPE, these should never be None
    assert proc.stdout is not None
    assert proc.stderr is not None

    stdout: str = ""
    if echo_output_to_stream is not None:
        # Print each line to the stream as it arrives
        while proc.poll() is None:
            line = proc.stdout.readline()
            echo_output_to_stream.write(line)
            stdout += line
    else:
        # Otherwise, simply wait for the command to finish and then grab stdout
        proc.wait()
        stdout = proc.stdout.read()

    stderr = proc.stderr.read()

    # Send to alkymi log
    log.debug(stdout)
    log.debug(stderr)

    # Always forward error messages to stderr if needed
    if echo_error_to_stream is not None and proc.returncode != 0:
        echo_error_to_stream.write(stderr)

    # Raise an error on failure
    if proc.returncode != 0:
        raise subprocess.CalledProcessError(proc.returncode, proc.args, stdout, stderr)
    return subprocess.CompletedProcess(proc.args, proc.returncode, stdout, stderr)



from collections import OrderedDict
from typing import Iterable, Callable, Optional, Tuple, Any, List, Dict, Union, cast, TypeVar
from itertools import chain

from . import checksums, serialization
from .logging import log
from .recipe import Recipe, CacheType, CleanlinessFunc
from .serialization import Output, OutputWithValue, CachedOutput

MappedInputs = Union[List[Any], Dict[Any, Any]]
MappedOutputs = Union[List[Output], Dict[Any, Output]]
MappedOutputsCached = Union[List[CachedOutput], Dict[Any, CachedOutput]]
MappedInputsChecksums = Union[List[Optional[str]], Dict[Any, Optional[str]]]

R = TypeVar("R")  # The return type of the bound function


class ForeachRecipe(Recipe[R]):
    """
    Special type of Recipe that applies its bound function to each input from a list or dictionary (similar to Python's
    built-in map() function). Evaluations of the bound function are cached and used to avoid reevaluation previously
    seen inputs, this means that changing the inputs to a ForeachRecipe may only trigger reevaluation of the bound
    function for some inputs, avoiding the overhead of recomputing things
    """

    def __init__(self, mapped_recipe: Recipe, ingredients: Iterable[Recipe], func: Callable[..., R], name: str,
                 transient: bool, cache: CacheType, cleanliness_func: Optional[CleanlinessFunc] = None):
        """
        Create a new ForeachRecipe

        :param mapped_recipe: A single Recipe to whose output (a list or dictionary) the bound function will be applied
                              to generate the new outputs (similar to Python's built-in map() function)
        :param ingredients: The dependencies of this Recipe - the outputs of these Recipes will be provided as arguments
                            to the bound function when called (following the item from the mapped_inputs sequence)
        :param func: The function to bind to this recipe
        :param name: The name of this Recipe
        :param transient: Whether to always (re)evaluate the created Recipe
        :param cache: The type of caching to use for this Recipe
        :param cleanliness_func: A function to allow a custom cleanliness check
        """
        self._mapped_inputs: Optional[MappedInputs] = None
        self._mapped_inputs_type: Optional[type] = None
        self._mapped_inputs_checksums: Optional[MappedInputsChecksums] = None
        self._mapped_inputs_checksum: Optional[str] = None
        self._mapped_outputs: Optional[MappedOutputs] = None
        self._mapped_outputs_checksum: Optional[str] = None
        super().__init__(func, chain([mapped_recipe], ingredients), name, transient, cache, cleanliness_func)

    @property
    def mapped_inputs_type(self) -> Optional[type]:
        return self._mapped_inputs_type

    @property
    def mapped_inputs_checksums(self) -> Optional[MappedInputsChecksums]:
        """
        :return: The computed checksums for the sequence of mapped inputs (this is set when mapped_inputs is set)
        """
        return self._mapped_inputs_checksums

    @property
    def mapped_inputs_checksum(self) -> Optional[str]:
        """
        :return: The summary of the mapped inputs checksum
        """
        return self._mapped_inputs_checksum

    @property
    def mapped_inputs(self) -> Optional[MappedInputs]:
        """
        :return: The sequence of inputs to apply the bound function to
        """
        return self._mapped_inputs

    @mapped_inputs.setter
    def mapped_inputs(self, mapped_inputs: MappedInputs) -> None:
        """
        Sets the sequence of inputs to apply the bound function to and computes the necessary checksums needed for
        checking dirtiness

        :param mapped_inputs: The sequence of inputs to apply the bound function to
        """
        if mapped_inputs is None:
            return

        # FIXME(mathias): This does unnecessary work by checksumming the same items multiple times during evaluation
        if isinstance(mapped_inputs, list):
            self._mapped_inputs_checksums = []
            for inp in mapped_inputs:
                self._mapped_inputs_checksums.append(checksums.checksum(inp))
        elif isinstance(mapped_inputs, dict):
            self._mapped_inputs_checksums = {}
            for key, inp in mapped_inputs.items():
                self._mapped_inputs_checksums[key] = checksums.checksum(inp)
        else:
            raise RuntimeError("Cannot handle mapped input of type: {}".format(type(mapped_inputs)))
        self._mapped_inputs = mapped_inputs
        self._mapped_inputs_type = type(self._mapped_inputs)

    @Recipe.outputs.getter  # type: ignore # see https://github.com/python/mypy/issues/1465
    def outputs(self) -> Optional[Union[Dict, List]]:
        """
        :return: The outputs of this ForeachRecipe in canonical form (None or a tuple with zero or more entries)
        """
        if self._mapped_outputs is None:
            return None
        if isinstance(self._mapped_outputs, list):
            return [output.value() for output in self._mapped_outputs]
        elif isinstance(self._mapped_outputs, dict):
            return {key: output.value() for key, output in self._mapped_outputs.items()}
        raise RuntimeError("Invalid type for mapped outputs")

    @property
    def output_checksum(self) -> Optional[str]:
        """
        :return: The computed checksums for the outputs (this is set when outputs are set)
        """
        if self._mapped_outputs_checksum is None:
            return None
        return self._mapped_outputs_checksum

    @property
    def mapped_outputs(self) -> Optional[MappedOutputs]:
        return self._mapped_outputs

    @property
    def mapped_outputs_checksums(self) -> Optional[MappedInputsChecksums]:
        """
        :return: The computed checksums for the sequence of mapped outputs
        """
        if self._mapped_outputs is None:
            return None
        if isinstance(self._mapped_outputs, list):
            return [output.checksum for output in self._mapped_outputs]
        elif isinstance(self._mapped_outputs, dict):
            return {key: output.checksum for key, output in self._mapped_outputs.items()}
        raise RuntimeError("Invalid type for mapped outputs")

    @property
    def outputs_valid(self) -> bool:
        """
        Check whether an output is still valid - this is currently only used to check files that may have been deleted
        or altered outside alkymi's cache. If no outputs have been produced yet, True will be returned.

        :return: Whether all outputs are still valid
        """
        if self._mapped_outputs is None:
            return True
        if isinstance(self._mapped_outputs, list):
            return all(output.valid for output in self._mapped_outputs)
        elif isinstance(self._mapped_outputs, dict):
            return all(output.valid for output in self._mapped_outputs.values())
        else:
            raise ValueError("Invalid type of mapped_outputs")

    def set_current_result(self, evaluated: MappedInputs, outputs: MappedOutputs, mapped_inputs_checksum: Optional[str],
                           other_input_checksums: Tuple[Optional[str], ...], completed: bool) -> None:
        """
        Stores the provided results in the recipe and caches them to disk if applicable

        :param evaluated: The inputs that were used to generate the provided outputs
        :param outputs: The outputs to store in this recipe
        :param mapped_inputs_checksum: The checksum of all mapped inputs
        :param other_input_checksums: The checksums of other (non-mapped) inputs
        :param completed: Bool indicating whether all mapped inputs have been processed
        """
        self.mapped_inputs = evaluated
        self._mapped_outputs = outputs
        self._mapped_outputs_checksum = checksums.checksum(outputs)
        self._last_function_hash = self.function_hash
        if completed:
            self._mapped_inputs_checksum = mapped_inputs_checksum
        else:
            # If not completed, use a dummy value to mark the inputs dirty
            self._mapped_inputs_checksum = "0xmissing_mapped_inputs_eval"
        self._input_checksums = (self._mapped_inputs_checksum,) + other_input_checksums
        self._save_state()

    def to_dict(self) -> OrderedDict:
        """
        :return: The ForeachRecipe as a dict for serialization purposes
        """
        # Force caching of all outputs (if they aren't already)
        serialized_mapped_outputs: Optional[Union[Dict, List]] = None
        if self._mapped_outputs is not None:
            if isinstance(self._mapped_outputs, list):
                outputs_list: List[CachedOutput] = []
                for output in self._mapped_outputs:
                    if isinstance(output, CachedOutput):
                        outputs_list.append(output)
                    elif isinstance(output, OutputWithValue):
                        outputs_list.append(serialization.cache(output, self.cache_path))
                    else:
                        raise RuntimeError("Output is of wrong type")
                serialized_mapped_outputs = [output.serialized for output in outputs_list]
                self._mapped_outputs = cast(List[Output], outputs_list)
            elif isinstance(self._mapped_outputs, dict):
                outputs_dict: Dict[Any, CachedOutput] = {}
                for key, output in self._mapped_outputs.items():
                    if isinstance(output, CachedOutput):
                        outputs_dict[key] = output
                    elif isinstance(output, OutputWithValue):
                        outputs_dict[key] = serialization.cache(output, self.cache_path)
                    else:
                        raise RuntimeError("Output is of wrong type")
                serialized_mapped_outputs = {key: output.serialized for key, output in outputs_dict.items()}
                self._mapped_outputs = cast(Dict[Any, Output], outputs_dict)

        return OrderedDict(
            name=self.name,
            input_checksums=self.input_checksums,
            mapped_outputs=serialized_mapped_outputs,
            mapped_outputs_checksums=self.mapped_outputs_checksums,
            mapped_outputs_checksum=self._mapped_outputs_checksum,
            output_checksum=self.output_checksum,
            last_function_hash=self._last_function_hash,
            mapped_inputs_checksums=self.mapped_inputs_checksums,
            mapped_inputs_checksum=self.mapped_inputs_checksum,
            mapped_type="dict" if self.mapped_inputs_type == dict else "list"
        )

    def restore_from_dict(self, old_state: Dict) -> None:
        """
        Restores the state of this ForeachRecipe from a previously cached state

        :param old_state: The old cached state to restore
        """
        log.debug("Restoring {} from dict".format(self._name))
        if old_state["input_checksums"] is not None:
            self._input_checksums = tuple(old_state["input_checksums"])
        mapped_type = old_state["mapped_type"]
        if mapped_type == "list":
            self._mapped_inputs_type = list
            self._mapped_outputs = [CachedOutput(None, checksum, serialized)
                                    for serialized, checksum in
                                    zip(old_state["mapped_outputs"], old_state["mapped_outputs_checksums"])]
        elif mapped_type == "dict":
            self._mapped_inputs_type = dict
            self._mapped_outputs = {key: CachedOutput(None, checksum, serialized)
                                    for (key, serialized), checksum in
                                    zip(old_state["mapped_outputs"].items(), old_state["mapped_outputs_checksums"])}
        else:
            raise ValueError("Unknown mapped type: {}".format(mapped_type))
        self._last_function_hash = cast(str, old_state["last_function_hash"])
        self._mapped_inputs_checksums = cast(MappedInputsChecksums, old_state["mapped_inputs_checksums"])
        self._mapped_inputs_checksum = cast(str, old_state["mapped_inputs_checksum"])
        self._mapped_outputs_checksum = cast(str, old_state["mapped_outputs_checksum"])



import inspect
from typing import Callable, TypeVar, Optional

from .config import CacheType
from .foreach_recipe import ForeachRecipe
from .recipe import Recipe

R = TypeVar("R")  # The return type of the bound function


def recipe(ingredients=(), name: Optional[str] = None, transient: bool = False, cache: CacheType = CacheType.Auto) -> \
        Callable[[Callable[..., R]], Recipe[R]]:
    """
    Convert a function into an alkymi Recipe to enable caching and conditional evaluation

    :param ingredients: The dependencies of this Recipe - the outputs of these Recipes will be provided as arguments to
                        the bound function when called in the order that they were provided. If not all arguments are
                        provided directly, alkymi will look up recipes that match the name of arguments automatically
    :param name: The name to assign to the created recipe - if not provided, the bound function's name will be used
    :param transient: Whether to always (re)evaluate the created Recipe
    :param cache: The type of caching to use for this Recipe
    :return: A callable that will yield the Recipe created from the bound function
    """
    ingredients = list(ingredients)
    num_provided_ingredients = len(ingredients)

    # Capture locals of calling scope to allow lookup of dependent Recipes in decorator
    outer_locals = inspect.stack(0)[1].frame.f_locals

    def _decorator(func: Callable[..., R]) -> Recipe[R]:
        """
        Closure to capture arguments from decorator

        :param func: The bound function to wrap in a Recipe
        :return: The created Recipe
        """
        # Find all the required arguments in the stored locals
        required_args = inspect.getfullargspec(func).args
        for arg_name in required_args[num_provided_ingredients:]:
            arg = outer_locals.get(arg_name, None)
            if arg is None:
                raise RuntimeError("Unable to find Recipe with name {} in enclosing scope".format(arg_name))
            if not isinstance(arg, Recipe):
                raise RuntimeError("Found argument with name {}, but not a Recipe".format(arg_name))
            ingredients.append(arg)

        recipe_name = func.__name__ if name is None else name
        return Recipe(func, ingredients, recipe_name, transient, cache)

    return _decorator


def foreach(mapped_inputs: Recipe, ingredients=(), name: Optional[str] = None, transient: bool = False,
            cache: CacheType = CacheType.Auto) -> \
        Callable[[Callable[..., R]], ForeachRecipe[R]]:
    """
    Convert a function into an alkymi Recipe to enable caching and conditional evaluation

    :param mapped_inputs: A single Recipe to whose output (a list or dictionary) the bound function will be applied to
                          generate the new outputs (similar to Python's built-in map() function)
    :param ingredients: The dependencies of this Recipe - the outputs of these Recipes will be provided as arguments to
                        the bound function when called in the order that they were provided. If not all arguments are
                        provided directly, alkymi will look up recipes that match the name of arguments automatically
    :param name: The name to assign to the created recipe - if not provided, the bound function's name will be used
    :param transient: Whether to always (re)evaluate the created Recipe
    :param cache: The type of caching to use for this Recipe
    :return: A callable that will yield the Recipe created from the bound function
    """
    ingredients = list(ingredients)
    num_provided_ingredients = len(ingredients)

    # Capture locals of calling scope to allow lookup of dependent Recipes in decorator
    outer_locals = inspect.stack(0)[1].frame.f_locals

    def _decorator(func: Callable[..., R]) -> ForeachRecipe[R]:
        """
        Closure to capture arguments from decorator

        :param func: The bound function to wrap in a ForeachRecipe
        :return: The created ForeachRecipe
        """
        # Find all the required arguments in the stored locals - ignore the first arg since that is the mapped arg
        required_args = inspect.getfullargspec(func).args[1:]
        for arg_name in required_args[num_provided_ingredients:]:
            arg = outer_locals.get(arg_name, None)
            if arg is None:
                raise RuntimeError("Unable to find Recipe with name {} in enclosing scope".format(arg_name))
            if not isinstance(arg, Recipe):
                raise RuntimeError("Found argument with name {}, but not a Recipe".format(arg_name))
            ingredients.append(arg)

        recipe_name = func.__name__ if name is None else name
        return ForeachRecipe(mapped_inputs, ingredients, func, recipe_name, transient, cache)

    return _decorator



import pickle
import re
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Optional, Any, Iterable, Union, Generator, Sequence, Dict, Type, TypeVar, Generic, cast

from . import checksums, AlkymiConfig

# TODO(mathias): This file needs to be reworked to be less complex/crazy. Some sort of class w/ recursive serialization
#                might help make this a lot more readable

# Shorthands for generator types used below
# Note that this doesn't fully represent the JSON hierarchy due to https://github.com/python/mypy/issues/731
CachePathGenerator = Generator[Path, None, None]
BaseSerializable = Union[str, int, float, None]
SerializableRepresentation = Union[BaseSerializable, Iterable[BaseSerializable], Dict[str, BaseSerializable]]

# Tokens used to signify to the deserializer func how to deserializer a given value
TOKEN_TEMPLATE = "!#{}#!"
TOKEN_REGEX = re.compile("(!#.*#!).*")


def create_token(name) -> str:
    """
    Creates a token using the 'TOKEN_TEMPLATE' to signify to the deserializer func how to deserializer a given value

    :param name: The name of the token
    :return: A new token for the given token name
    """
    return TOKEN_TEMPLATE.format(name)


# Create tokens
PATH_TOKEN = create_token("path")
PICKLE_TOKEN = create_token("pickle")
BYTES_TOKEN = create_token("bytes")

S = TypeVar("S")  # The type that a Serializer subclass acts on


class Serializer(Generic[S]):
    """
    Abstract base class for classes that enable serialization/deserialization of classes not in the standard library
    """

    @staticmethod
    def serialize(value: S, cache_path: Path) -> str:
        raise NotImplementedError()

    @staticmethod
    def deserialize(path: Path) -> S:
        raise NotImplementedError()


# Load additional serializers and deserializers based on available libs
additional_serializers: Dict[Any, Type[Serializer]] = {}
additional_deserializers: Dict[str, Type[Serializer]] = {}
try:
    import numpy as np  # NOQA


    class NdArraySerializer(Serializer[np.ndarray]):
        """
        Numpy array serializer/deserializer
        """
        TOKEN = create_token("ndarray")

        @staticmethod
        def serialize(array: np.ndarray, cache_path: Path) -> str:
            """
            Saves a numpy array to an .npy file

            :param array: The file so save
            :param cache_path: The path to save the array to
            :return: A tokenized path to allow the deserializer to call the "deserialize" function of this module
            """
            path = cache_path.with_suffix(".npy")
            np.save(str(path), array, allow_pickle=False)
            return "{}{}".format(NdArraySerializer.TOKEN, path)

        @staticmethod
        def deserialize(path: Path) -> np.ndarray:
            """
            Loads the numpy .npy file from the provided path

            :param path: The path to the numpy file to load
            :return: The decoded numpy array
            """
            return np.load(str(path), allow_pickle=False)


    additional_serializers[np.ndarray] = NdArraySerializer
    additional_deserializers[NdArraySerializer.TOKEN] = NdArraySerializer
except ImportError:
    pass


def serialize_item(item: Any, cache_path_generator: CachePathGenerator) -> SerializableRepresentation:
    """
    Serializes an item (potentially recursively)

    :param item: The item to serialize (may be nested)
    :param cache_path_generator: The generator to use for generating cache paths
    :return: The serialized item
    """
    if item is None:
        return None

    itype = type(item)
    serializer = additional_serializers.get(itype, None)
    if serializer is not None:
        return serializer.serialize(item, next(cache_path_generator))
    elif isinstance(item, Path):
        # External path - store the checksum of the file at the current point in time
        file_checksum = checksums.checksum(item)
        return "{}{}:{}".format(PATH_TOKEN, file_checksum, item)
    elif itype in (str, float, int, bool):
        return item
    elif isinstance(item, bytes):
        output_file = next(cache_path_generator)
        with output_file.open("wb") as f:
            f.write(item)
        return "{}{}".format(BYTES_TOKEN, output_file)
    elif isinstance(item, Sequence):
        # recursive types are not supported by mypy yet
        return [serialize_item(subitem, cache_path_generator) for subitem in item]  # type: ignore
    elif isinstance(item, dict):
        keys = serialize_item(list(item.keys()), cache_path_generator)
        values = serialize_item(list(item.values()), cache_path_generator)
        return dict(keys=keys, values=values)
    else:
        # As a last resort, try to dump as pickle
        if not AlkymiConfig.get().allow_pickling:
            raise RuntimeError("Pickling disabled - cannot serialize type: {}".format(type(item)))

        try:
            output_file = next(cache_path_generator)
            with output_file.open("wb") as f:
                pickle.dump(item, f, protocol=pickle.HIGHEST_PROTOCOL)
            return "{}{}".format(PICKLE_TOKEN, output_file)
        except pickle.PicklingError:
            raise Exception("Cannot serialize item of type: {}".format(type(item)))


def deserialize_item(item: SerializableRepresentation) -> Any:
    """
    Deserializes an item (potentially recursively)

    :param item: The item to deserialize (may be nested)
    :return: The deserialized item
    """
    if item is None:
        return None

    if isinstance(item, str):
        m = re.match(TOKEN_REGEX, item)
        if m is None:
            # Regular string
            return item
        else:
            if item.startswith(PATH_TOKEN):
                # Path encoded as string with checksum, e.g. "!#path#!CHECKSUM_HERE:/what/a/path"
                non_token_part = item[len(PATH_TOKEN):]
                _, path_str = non_token_part.split(":", maxsplit=1)
                return Path(path_str)
            elif item.startswith(BYTES_TOKEN):
                # Bytes dumped to file
                with open(item[len(BYTES_TOKEN):], "rb") as f:
                    return f.read()
            elif item.startswith(PICKLE_TOKEN):
                # Arbitrary object encoded as pickle
                if not AlkymiConfig.get().allow_pickling:
                    raise RuntimeError("Pickling disabled - cannot deserialize item: {}".format(item))
                with open(item[len(PICKLE_TOKEN):], "rb") as f:
                    return pickle.loads(f.read())
            else:
                found_token = m.group(1)
                deserializer = additional_deserializers.get(found_token, None)
                if deserializer is not None:
                    return deserializer.deserialize(Path(item[len(found_token):]))
                else:
                    raise RuntimeError("No deserializer found for token: {}".format(found_token))
    elif isinstance(item, float) or isinstance(item, int):
        return item
    elif isinstance(item, Sequence):
        return list(deserialize_item(subitem) for subitem in item)
    elif isinstance(item, dict):
        # These should never be triggered, because we always store keys and values as lists in serialize_item(), but
        # this makes the type-checker happy
        if not isinstance(item["keys"], Iterable):
            raise ValueError("'keys' entry must be a list")
        if not isinstance(item["values"], Iterable):
            raise ValueError("'values' entry must be a list")
        keys = list(deserialize_item(key) for key in item["keys"])
        values = list(deserialize_item(val) for val in item["values"])
        return {key: value for key, value in zip(keys, values)}
    else:
        raise Exception("Cannot deserialize item of type: {}".format(type(item)))


def is_valid_serialized(item: SerializableRepresentation) -> bool:
    """
    Recursively check validity of a serialized representation. Currently just looks for external files represented by
    Path objects, and then compares the stored checksum of each such item with the current checksum (computed from the
    current file contents)

    :param item: The serialized representation to check validity for
    :return: True if the input is still valid
    """
    if isinstance(item, str):
        if item.startswith(PATH_TOKEN):
            # Path encoded as string with checksum, e.g. "!#path#!CHECKSUM_HERE:/what/a/path"
            non_token_part = item[len(PATH_TOKEN):]
            stored_checksum, path_str = non_token_part.split(":", maxsplit=1)
            current_checksum = checksums.checksum(Path(path_str))
            return stored_checksum == current_checksum
    elif isinstance(item, Sequence):
        return all(is_valid_serialized(subitem) for subitem in item)

    # Other types are always valid
    return True


T = TypeVar('T')


class Output(Generic[T], metaclass=ABCMeta):
    """
    Abstract base class for keeping track of outputs of Recipes
    """

    def __init__(self, checksum: str):
        """
        Create a new Output and assign the checksum

        :param checksum: The checksum for the object
        """
        self._checksum = checksum

    @property
    def checksum(self) -> str:
        """
        :return: The checksum of this Output
        """
        return self._checksum

    @property
    def valid(self) -> bool:
        """
        :return: Whether this Output is still valid (e.g. an external file pointed to by a Path can have been altered)
        """
        return NotImplemented

    @abstractmethod
    def value(self) -> T:
        """
        :return: The value associated with this Output
        """
        pass


class OutputWithValue(Output):
    """
    An Output that is guaranteed to have an in-memory value - all outputs start out as this before being cached
    """

    def __init__(self, value: T, checksum: str):
        """
        Create a new OutputWithValue

        :param value: The value of the output
        :param checksum: The checksum of the output
        """
        super().__init__(checksum)
        self._value = value

    @Output.valid.getter  # type: ignore # see https://github.com/python/mypy/issues/1465
    def valid(self) -> bool:
        if self._value is None:
            return True
        # TODO(mathias): Find out if this is too expensive in general
        return checksums.checksum(self._value) == self.checksum

    def value(self) -> T:
        return self._value


class CachedOutput(Output):
    """
    An Output that has been cached - may or may not have it's associated value in-memory
    """

    def __init__(self, value: Optional[T], checksum: str, serializable_representation: SerializableRepresentation):
        """
        Create a new CachedOutput

        :param value: The value of the output
        :param checksum: The checksum of the output
        :param serializable_representation: The serialized representation of the output
        """
        super().__init__(checksum)
        self._value = value
        self._serializable_representation = serializable_representation

    @Output.valid.getter  # type: ignore # see https://github.com/python/mypy/issues/1465
    def valid(self) -> bool:
        return is_valid_serialized(self._serializable_representation)

    def value(self) -> T:
        # Deserialize the value if it isn't already in memory
        if self._value is None:
            value = from_cache(self._serializable_representation)
            self._value = cast(T, value)
        return self._value

    @property
    def serialized(self) -> SerializableRepresentation:
        """
        :return: A serializable representation of the value of this output
        """
        return self._serializable_representation


def cache(output: OutputWithValue, base_path: Path) -> CachedOutput:
    """
    Cache an in-memory OutputWithValue, thus converting it to a CachedOutput. The resulting output will retain the value
    in-memory

    :param output: The Output to cache
    :param base_path: The directory to use for this serialization. A subdirectory will be created to store complex
        serialized objects
    :return: The cached output
    """
    value = output.value()  # type: ignore  # Make all Output types fully generic for this to work
    checksum = output.checksum
    if value is None:
        return CachedOutput(value, checksum, None)

    cache_path = base_path / checksum
    cache_path.mkdir(exist_ok=True)

    def cache_path_generator() -> Generator[Path, None, None]:
        i = 0
        while True:
            yield cache_path / str(i)
            i += 1

    return CachedOutput(value, checksum, serialize_item(value, cache_path_generator()))


def from_cache(serializable_representation: SerializableRepresentation) -> Any:
    """
    Deserialize an output from the cache using its serialized representation

    :param serializable_representation: The serialized representation to deserialize
    :return: The deserialized object
    """
    return deserialize_item(serializable_representation)



from .config import AlkymiConfig  # NOQA
from .lab import Lab  # NOQA
from .logging import log  # NOQA
from .decorators import recipe, foreach  # NOQA
from .recipes import glob_files, file, arg  # NOQA
from .utils import call  # NOQA
from .types import Status  # NOQA
from . import checksums  # NOQA
from . import core  # NOQA

# Define version
from .version import VERSION, __version__  # NOQA



from pathlib import Path
from typing import List, Tuple, Any, Dict, Iterable, Union, TypeVar, Type, Optional

from .config import CacheType
from .recipe import Recipe


def glob_files(name: str, directory: Path, pattern: str, recursive: bool, cache=CacheType.Auto) -> Recipe[List[Path]]:
    """
    Create a Recipe that will glob files in a directory and return them as a list. The created recipe will only be
    considered dirty if the file paths contained in the glob changes (not if the contents of any one file changes)

    :param name: The name to give the created Recipe
    :param directory: The directory in which to perform the glob
    :param pattern: The pattern to use for the glob, e.g. `*.py`
    :param recursive: Whether to glob recursively into subdirectories
    :param cache: The type of caching to use for this Recipe
    :return: The created Recipe
    """

    def _glob_recipe() -> List[Path]:
        """
        The bound function that performs the glob

        :return: The glob of files
        """
        if recursive:
            return list(directory.rglob(pattern))
        else:
            return list(directory.glob(pattern))

    def _check_clean(last_outputs: List[Path]) -> bool:
        """
        If rerunning glob produces the same list of files, then the recipe is clean

        :param last_outputs: The last set of outputs created by this Recipe
        :return: True if clean
        """
        if last_outputs is None:
            return False
        return _glob_recipe() == last_outputs

    return Recipe(_glob_recipe, [], name, transient=False, cache=cache, cleanliness_func=_check_clean)


def file(name: str, path: Path, cache=CacheType.Auto) -> Recipe[Path]:
    """
    Create a Recipe that outputs a single file

    :param name: The name to give the created Recipe
    :param path: The path to the file to output
    :param cache: The type of caching to use for this Recipe
    :return: The created Recipe
    """
    # Reference the path as a string to avoid changes to the Path object to change the checksum of the function
    path_as_str = str(path)

    def _file_recipe() -> Path:
        return Path(path_as_str)

    return Recipe(_file_recipe, [], name, transient=False, cache=cache)


def zip_results(name: str, recipes: Iterable[Recipe], cache=CacheType.Auto) \
        -> Recipe[Union[List[Tuple[Any, ...]], Dict[Any, Tuple[Any, ...]]]]:
    """
    Create a Recipe that zips the outputs from a number of recipes into elements, similar to Python's built-in zip().
    Notably, dictionaries are handled a bit differently, in that a dictionary is returned with keys mapping to tuples
    from the different inputs, i.e.::

        {"1": 1} zip {"1", "one"} -> {"1", (1, "one")}

    :param name: The name to give the created Recipe
    :param recipes: The recipes to zip. These must return lists or dictionaries
    :param cache: The type of caching to use for this Recipe
    :return: The created Recipe
    """

    def _zip_results(*iterables: Union[List, Dict]) \
            -> Union[List[Tuple[Any, ...]], Dict[Any, Tuple[Any, ...]]]:
        # Sanity checks
        if not iterables or len(iterables) == 0:
            return []

        if any(not isinstance(iterable, Iterable) for iterable in iterables):
            raise ValueError("Cannot zip non-iterable inputs")

        first_iterable = iterables[0]
        if any(not isinstance(iterable, type(first_iterable)) for iterable in iterables):
            raise ValueError("Cannot zip inputs of different types")

        num_items = len(first_iterable)
        if any(len(iterable) != num_items for iterable in iterables):
            raise ValueError("Cannot zip inputs of different length")

        # Handle the actual zipping operation
        if isinstance(first_iterable, list):
            return list(zip(*iterables))
        elif isinstance(first_iterable, dict):
            return {
                key: tuple(iterable[key] for iterable in iterables)
                for key in first_iterable.keys()
            }
        else:
            raise ValueError("Type: {} not supported in _zip_results()".format(type(first_iterable)))

    return Recipe(_zip_results, recipes, name, transient=False, cache=cache)


T = TypeVar('T')


class Arg(Recipe[T]):
    """
    Class providing a stateful argument

    To use, create an Arg instance with the initial value for your argument, e.g. ``Arg(0)``, then provide as a
    recipe to downstream recipes. To change the input arguments, call ``set()`` again - this will mark the recipe as
    dirty and cause reevaluation of downstream recipe(s)
    """

    def __init__(self, arg: T, name: str, cache=CacheType.Auto):
        """
        Create a new Arg instance with initial argument value

        :param arg: The initial argument value
        :param name: The name to give the created Recipe
        :param cache: The type of caching to use for this Recipe
        """
        self._arg = arg
        self._type = type(arg)
        self._subtype = next((type(val) for val in arg), None) if isinstance(arg, Iterable) else None
        super().__init__(self._produce_arg, [], name, transient=False, cache=cache, cleanliness_func=self._clean)

    @property
    def type(self) -> Type[T]:
        """
        :return: The type of the argument
        """
        return self._type

    @property
    def subtype(self) -> Optional[Any]:
        """
        :return: The subtype of the argument (e.g. the type of items contained in a list). Will be None for non-iterable
                 types
        """
        return self._subtype

    def _produce_arg(self) -> T:
        """
        :return: The current argument
        """
        return self._arg

    def _clean(self, last_outputs: Any) -> bool:
        """
        Checks whether the argument has changed since the last evaluation

        :param last_outputs: The last set of outputs (argument)
        :return: True if the argument remain the same
        """
        return self._arg == last_outputs

    def set(self, _arg: T) -> None:
        """
        Change the argument, causing the recipe to need reevaluation

        :param _arg: The new argument
        """
        if not isinstance(_arg, self._type):
            raise TypeError("Type of argument ({}) not {}".format(type(_arg), self._type))
        self._arg = _arg


def arg(_arg: T, name: str, cache=CacheType.Auto) -> Arg[T]:
    """
    Shorthand for creating an ``Arg`` instance

    :param _arg: The initial argument to use
    :param name: The name to give the created Recipe
    :param cache: The type of caching to use for this Recipe
    :return: The created ``Arg`` instance
    """
    return Arg(_arg, name=name, cache=cache)



import setuptools

with open("README.md", encoding='utf-8') as f:
    long_description = f.read()

version = {}
with open('alkymi/version.py') as fp:
    exec(fp.read(), version)

setuptools.setup(
    name="alkymi",
    description="alkymi - Pythonic task automation",
    version=version['__version__'],
    license="MIT",
    author="Mathias Bøgh Stokholm",
    author_email="mathias.stokholm@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MathiasStokholm/alkymi",
    packages=["alkymi"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Scientific/Engineering",
    ],
    keywords=["automation", "pipeline", "validation", "preprocessing", "make", "build", "task"],
    project_urls={
        "Source": "https://github.com/MathiasStokholm/alkymi/",
        "Tracker": "https://github.com/MathiasStokholm/alkymi/issues",
        "Documentation": "https://alkymi.readthedocs.io/en/latest/",
    },
    python_requires=">=3.6",
    install_requires=[
        "networkx>=2.0"
    ],
    extras_require={
        "xxhash": ["xxhash>=2.0.0"]
    },
    package_data={
        "alkymi": ["py.typed"],
    },
)



import enum
from pathlib import Path
import os
from typing import Optional


@enum.unique
class CacheType(enum.Enum):
    """
    Supported caching mechanisms
    """
    Cache = 0  # Enable caching
    NoCache = 1  # Disable caching
    Auto = 2  # Enable or disable caching based on AlkymiConfig.cache setting


@enum.unique
class FileChecksumMethod(enum.Enum):
    """
    Supported ways of calculating the checksum for a Path object representing a file
    """
    HashContents = 0  # Hash the contents of the file
    ModificationTimestamp = 1  # Use timestamp of last modification


class AlkymiConfig:
    """
    Global singleton config for alkymi
    """
    __instance: Optional["AlkymiConfig"] = None

    @staticmethod
    def get() -> 'AlkymiConfig':
        """
        :return: The singleton instance of AlkymiConfig
        """
        if AlkymiConfig.__instance is None:
            AlkymiConfig()
        assert AlkymiConfig.__instance is not None
        return AlkymiConfig.__instance

    def __init__(self):
        """
        Private initializer that creates and sets the singleton instance
        """
        if AlkymiConfig.__instance is not None:
            raise Exception("This class is a singleton!")

        # Set default values in config
        AlkymiConfig.__instance = self
        self._cache = True
        self._cache_path = None
        self._allow_pickling = True
        self._file_checksum_method = FileChecksumMethod.HashContents

    @property
    def cache(self) -> bool:
        """
        :return: Whether to enable alkymi caching globally (see CacheType.Auto)
        """
        return self._cache

    @cache.setter
    def cache(self, enable_cache: bool) -> None:
        """
        Set whether to enable alkymi caching globally (see CacheType.Auto)

        :param enable_cache: Whether to enable alkymi caching globally
        """
        self._cache = enable_cache

    @property
    def cache_path(self) -> Optional[Path]:
        """
        :return: A user-provided location to place the cache
        """
        return self._cache_path

    @cache_path.setter
    def cache_path(self, cache_path: Path) -> None:
        """
        Set a custom cache path to override the default caching location

        :param cache_path: The custom location to place the cache
        """
        if not cache_path.is_dir():
            raise ValueError("Cache path '{}' must be a directory".format(cache_path))
        if not os.access(str(cache_path), os.W_OK):
            raise ValueError("Cache path '{}' must be writeable".format(cache_path))
        self._cache_path = cache_path

    @property
    def allow_pickling(self) -> bool:
        """
        :return: Whether to allow pickling for serialization, deserialization and checksumming
        """
        return self._allow_pickling

    @allow_pickling.setter
    def allow_pickling(self, allow_pickling: bool) -> None:
        """
        Set whether to allow pickling for serialization, deserialization and checksumming

        :param allow_pickling: Whether to allow pickling for serialization, deserialization and checksumming
        """
        self._allow_pickling = allow_pickling

    @property
    def file_checksum_method(self) -> FileChecksumMethod:
        """
        :return: The currently used method for calculating file checksums (for Path objects)
        """
        return self._file_checksum_method

    @file_checksum_method.setter
    def file_checksum_method(self, file_checksum_method: FileChecksumMethod) -> None:
        """
        Set the method to use for calculating file checksums (for Path objects)

        :param file_checksum_method: The method to use for calculating file checksums (for Path objects)
        """
        self._file_checksum_method = file_checksum_method


# Force creation of singleton
ALKYMI_CONFIG = AlkymiConfig.get()



import asyncio
import concurrent.futures
from asyncio import Future, AbstractEventLoop
from typing import Dict, Tuple, Optional, Awaitable, Any

from . import checksums
from .foreach_recipe import ForeachRecipe, MappedOutputs, MappedInputs
from .serialization import OutputWithValue
from .types import Status
from .recipe import Recipe, R
from .logging import log

import networkx as nx

OutputsAndChecksums = Tuple[R, Optional[str]]


def create_graph(recipe: Recipe[R]) -> nx.DiGraph:
    """
    Create a Directed Acyclic Graph (DAG) based on the provided recipe
    Each node in the graph represents a recipe, and has an associated "status" attribute

    :param recipe: The recipe to construct a graph for
    :return: The constructed graph
    """
    log.debug(f'Building graph for {recipe.name}')
    graph = nx.DiGraph()
    _add_recipe_to_graph(recipe, graph)
    return graph


def _add_recipe_to_graph(recipe: Recipe, graph: nx.DiGraph) -> None:
    """
    Add a node representing a recipe to the graph, and recursively add dependencies of the provided recipe as nodes with
    edges to the current node

    :param recipe: The recipe to create a node for
    :param graph: The graph to which the new node and edges should be added
    """
    graph.add_node(recipe)
    log.debug(f'Added {recipe.name} to graph')

    # For each ingredient, add an edge from the ingredient to this recipe
    for _ingredient in recipe.ingredients:
        _add_recipe_to_graph(_ingredient, graph)
        graph.add_edge(_ingredient, recipe)


def _compute_status(recipe: Recipe, graph: nx.DiGraph, statuses: Dict[Recipe, Status]) -> Status:
    """
    Compute the status for the provided recipe (and recursively for dependencies) and add them all to the provided
    'statuses' dict

    :param recipe: The recipe to compute the status for
    :param graph: The graph to use for establishing statuses
    :param statuses: A dictionary used to collect the results of the recursive call
    :return status: The status of this recipe
    """

    def _store_and_return(_status: Status) -> Status:
        """
        Helper function to store a result in the 'statuses' dict before returning it
        """
        statuses[recipe] = _status
        return _status

    # Determine statuses of dependencies - note that this recursively computes statuses for dependencies and adds them
    # to 'statuses', so this needs to be called before any return
    dependencies = tuple(graph.predecessors(recipe))
    ingredient_statuses = tuple(
        _compute_status(ingredient, graph, statuses)
        for ingredient in dependencies
    )

    # If output checksum is None (or transient), a full re-evaluation is needed
    if recipe.transient or recipe.output_checksum is None:
        return _store_and_return(Status.NotEvaluatedYet)

    # Check if one or more ingredients (dependencies) are dirty
    if any(status != Status.Ok for status in ingredient_statuses):
        return _store_and_return(Status.IngredientDirty)

    # Check if one or more ingredients (dependencies) are dirty
    ingredient_output_checksums: Tuple[Optional[str], ...] = tuple(
        ingredient.output_checksum
        for ingredient in dependencies
        if ingredient.output_checksum is not None
    )
    status = is_clean(recipe, ingredient_output_checksums)
    return _store_and_return(status)


def compute_recipe_status(recipe: Recipe[R], graph: nx.DiGraph) -> Dict[Recipe, Status]:
    """
    Compute the Status for the provided recipe and all dependencies (ingredients or mapped inputs)

    :param recipe: The recipe for which status should be computed
    :param graph: The graph representing the recipe and all its dependencies
    :return: The status of the provided recipe and all dependencies as a dictionary
    """
    # Start recursion with an empty status dict
    statuses: Dict[Recipe, Status] = {}
    _compute_status(recipe, graph, statuses)
    return statuses


def invoke(recipe: Recipe, inputs: Tuple[Any, ...], input_checksums: Tuple[Optional[str], ...]) -> OutputsAndChecksums:
    """
    Evaluate the Recipe using the provided inputs. This will call the bound function on the inputs.

    :param recipe: The recipe to evaluate given the provided inputs
    :param inputs: The inputs provided by the ingredients (dependencies) of the Recipe
    :param input_checksums: The (possibly new) input checksum
    """
    log.debug('Invoking recipe: {}'.format(recipe.name))
    outputs = recipe(*inputs)
    recipe.set_result(outputs, input_checksums)
    return recipe.outputs, recipe.output_checksum


def invoke_foreach(recipe: ForeachRecipe, inputs: Tuple[Any, ...],
                   input_checksums: Tuple[Optional[str], ...],
                   executor: Optional[concurrent.futures.Executor]) -> OutputsAndChecksums:
    """
    Evaluate the ForeachRecipe using the provided inputs. This will apply the bound function to each item in the
    "mapped_inputs". If the result for any item is already cached, that result will be used instead (the checksum
    is used to check this). Only items from the immediately previous invoke call will be cached

    :param recipe: The ForeachRecipe to evaluate given the provided inputs
    :param inputs: The inputs provided by the ingredients (dependencies) of the ForeachRecipe
    :param input_checksums: The (possibly new) input checksum to use for checking cleanliness
    :param executor: The executor to use for calling the bound function in parallel
    """
    log.debug("Invoking recipe: {}".format(recipe.name))

    # The first ingredient will provide the sequence to apply the bound function too
    mapped_inputs = inputs[0]
    mapped_inputs_checksum = input_checksums[0]
    other_inputs = inputs[1:]
    other_input_checksums = input_checksums[1:]

    if not (isinstance(mapped_inputs, list) or isinstance(mapped_inputs, dict)):
        raise RuntimeError("Cannot handle type in invoke(): {}".format(type(mapped_inputs)))

    mapped_inputs_of_same_type = recipe.mapped_inputs_type == type(mapped_inputs)

    # Check if a full reevaluation across all mapped inputs is needed
    needs_full_eval = recipe.transient or not mapped_inputs_of_same_type

    # Check if bound function has changed - this should cause a full reevaluation
    if not needs_full_eval:
        if recipe.function_hash != recipe.last_function_hash:
            needs_full_eval = True

    # Check if we actually need to do any work (in case everything remains the same as last invocation)
    if not needs_full_eval and recipe.input_checksums is not None:
        if recipe.input_checksums == input_checksums:
            # Outputs have to be valid for us to return them
            if recipe.outputs_valid:
                log.debug("Returning early since mapped inputs did not change since last evaluation")
                return recipe.outputs, recipe.output_checksum
        # If input checksums do not match, a full re-evaluation is needed if the mapped checksums match, since the
        # mismatch has to be caused by a non-mapped input
        elif mapped_inputs_checksum == recipe.mapped_inputs_checksum:
            needs_full_eval = True

    # Catch up on already done work
    # TODO(mathias): Refactor this insanity to avoid the list/dict type checking
    outputs: MappedOutputs = [] if isinstance(mapped_inputs, list) else {}
    evaluated: MappedInputs = [] if isinstance(mapped_inputs, list) else {}
    not_evaluated: MappedInputs = [] if isinstance(mapped_inputs, list) else {}
    if needs_full_eval or recipe.mapped_outputs is None:
        not_evaluated = mapped_inputs
    else:
        if isinstance(mapped_inputs, list) and isinstance(outputs, list) \
                and isinstance(evaluated, list) and isinstance(not_evaluated, list):
            for item in mapped_inputs:
                # Try to look up cached result for this input
                try:
                    new_checksum = checksums.checksum(item)
                    idx = recipe.mapped_inputs_checksums.index(new_checksum)  # type: ignore
                    found_checksum = recipe.mapped_inputs_checksums[idx]  # type: ignore
                    if new_checksum == found_checksum:
                        found_output = recipe.mapped_outputs[idx]
                        if found_output.valid:
                            outputs.append(found_output)
                            evaluated.append(item)
                            continue
                except ValueError:
                    pass
                not_evaluated.append(item)
        elif isinstance(mapped_inputs, dict):
            for key, item in mapped_inputs.items():
                # Try to look up cached result for this input
                found_checksum = recipe.mapped_inputs_checksums.get(key, None)  # type: ignore
                if found_checksum is not None:
                    new_checksum = checksums.checksum(key)
                    if new_checksum == found_checksum:
                        found_output = recipe.mapped_outputs[key]
                        if found_output.valid:
                            outputs[key] = found_output
                            evaluated[key] = item
                            continue
                not_evaluated[key] = item

    log.debug("Num already cached results: {}/{}".format(len(evaluated), len(mapped_inputs)))
    if len(evaluated) == len(mapped_inputs):
        log.debug("Returning early since all items were already cached")
        recipe.set_current_result(evaluated, outputs, mapped_inputs_checksum, other_input_checksums, True)
        return recipe.outputs, recipe.output_checksum

    # Perform remaining work - store state every time an evaluation is successful
    if isinstance(not_evaluated, list) and isinstance(outputs, list) and isinstance(evaluated, list):
        if executor is not None:
            results = executor.map(lambda _item: recipe(_item, *other_inputs), not_evaluated)
        else:
            results = map(lambda _item: recipe(_item, *other_inputs), not_evaluated)
        for item, result in zip(not_evaluated, results):
            outputs.append(OutputWithValue(result, checksums.checksum(result)))
            evaluated.append(item)
            recipe.set_current_result(evaluated, outputs, mapped_inputs_checksum, other_input_checksums, False)
    elif isinstance(not_evaluated, dict):
        if executor is not None:
            results = executor.map(lambda _item: recipe(_item, *other_inputs), not_evaluated.values())
        else:
            results = map(lambda _item: recipe(_item, *other_inputs), not_evaluated.values())
        for (key, item), result in zip(not_evaluated.items(), results):
            outputs[key] = OutputWithValue(result, checksums.checksum(result))
            evaluated[key] = item
            recipe.set_current_result(evaluated, outputs, mapped_inputs_checksum, other_input_checksums, False)

    recipe.set_current_result(evaluated, outputs, mapped_inputs_checksum, other_input_checksums, True)
    return recipe.outputs, recipe.output_checksum


def retrieve_recipe_outputs(foreach_executor: Optional[concurrent.futures.Executor], recipe: Recipe, status: Status,
                            inputs_and_checksums: Tuple[OutputsAndChecksums, ...] = ()) -> OutputsAndChecksums:
    """
    Helper function to delegate recipe invocation calls

    :param foreach_executor: The executor (if any) that should be used for evaluating ForeachRecipes in parallel
    :param recipe: The recipe to evaluate using the executor
    :param status: The status of the recipe - used to skip evaluation if unnecessary
    :param inputs_and_checksums: The inputs and checksums of these to provide to the recipe for evaluation
    :return: The output(s) and checksum(s) of the recipe
    """
    # If status is not Ok, call invoke to run recipe
    if status != Status.Ok:
        _inputs = tuple(inp[0] for inp in inputs_and_checksums)
        _input_checksums = tuple(inp[1] for inp in inputs_and_checksums)
        if isinstance(recipe, ForeachRecipe):
            return invoke_foreach(recipe, _inputs, _input_checksums, foreach_executor)
        else:
            return invoke(recipe, _inputs, _input_checksums)
    return recipe.outputs, recipe.output_checksum


async def schedule(loop: AbstractEventLoop, graph_executor: concurrent.futures.Executor,
                   foreach_executor: Optional[concurrent.futures.Executor], recipe: Recipe,
                   statuses: Dict[Recipe, Status],
                   inputs_and_checksum_futures: Tuple[Awaitable[OutputsAndChecksums], ...]) -> Future:
    """
    Helper function used to asynchronously await inputs from dependant recipe futures, and then retrieve the output of
    the provided recipe using the provided executor (evaluating it if necessary)

    :param loop: The asyncio event loop to use for scheduling the recipe evaluation
    :param graph_executor: The executor that is being used to evaluate the graph
    :param foreach_executor: The executor (if any) that should be used for evaluating ForeachRecipes in parallel
    :param recipe: The recipe to evaluate using the executor
    :param statuses: The statuses of the recipes contained in the graph - used to skip evaluation if unnecessary
    :param inputs_and_checksum_futures: A tuple of futures to await before providing them to the recipe evaluation
    :return: A future that will eventually return the output(s) and checksum(s) of the recipe
    """
    inputs_and_checksums = tuple([await inp for inp in inputs_and_checksum_futures])
    return loop.run_in_executor(graph_executor, retrieve_recipe_outputs, foreach_executor, recipe, statuses[recipe],
                                inputs_and_checksums)


def evaluate_recipe(recipe: Recipe[R], graph: nx.DiGraph, statuses: Dict[Recipe, Status], jobs: int) -> \
        OutputsAndChecksums[R]:
    """
    Evaluate a Recipe, including any dependencies that are not up-to-date

    :param recipe: The recipe to evaluate
    :param graph: The graph to use for evaluation
    :param statuses: The statuses of the recipes contained in the graph - used to skip evaluation if unnecessary
    :param jobs: The number of jobs to use for evaluating the recipe in parallel, 1 job corresponds to no parallelism,
                 zero or negative values will cause alkymi to use the system's default number of jobs
    :return: The output(s) and checksum(s) of the evaluated recipe
    """
    # Create the executor to use for evaluating recipes
    executor = concurrent.futures.ThreadPoolExecutor(max_workers=jobs if jobs > 0 else None)

    # If we have more than 1 job (threads) available, also use the executor for running ForeachRecipes in parallel
    foreach_executor = executor if jobs > 1 else None

    # Create the asyncio event loop and set it on the calling thread
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    async def _execute() -> OutputsAndChecksums[R]:
        # Sort the graph topographically, such that any recipe in the sorted list only depends on earlier recipes
        # This guarantees that futures only depend on already created futures
        recipes = list(nx.topological_sort(graph))

        # Schedule all recipes to execute as soon as their inputs are available
        tasks: Dict[Recipe, Future] = {}
        for _recipe in recipes:
            input_futures = tuple(tasks[ingredient] for ingredient in _recipe.ingredients)
            tasks[_recipe] = await schedule(loop, executor, foreach_executor, _recipe, statuses, input_futures)

        # Wait for future for target recipe to return
        return await tasks[recipe]

    # Return the output and checksum of the final recipe
    output, checksum = loop.run_until_complete(_execute())
    return output, checksum


def is_clean(recipe: Recipe[R], new_input_checksums: Tuple[Optional[str], ...]) -> Status:
    """
    Check whether a Recipe is clean (result is cached) based on a set of (potentially new) input checksums

    :param recipe: The Recipe to check for cleanliness
    :param new_input_checksums: The (potentially new) input checksums to use for checking cleanliness
    :return: Whether the recipe is clean represented by the Status enum
    """
    # Non-pure function may have been changed by external circumstances, use custom check
    if recipe.custom_cleanliness_func is not None:
        if not recipe.custom_cleanliness_func(recipe.outputs):
            return Status.CustomDirty

    # Not clean if outputs were never generated
    if recipe.output_checksum is None:
        return Status.NotEvaluatedYet

    # Compute input checksums and perform equality check
    if recipe.input_checksums != new_input_checksums:
        log.debug('{} -> dirty: input checksums changed'.format(recipe.name))
        return Status.InputsChanged

    # Check if bound function has changed
    if recipe.last_function_hash is not None:
        if recipe.last_function_hash != recipe.function_hash:
            return Status.BoundFunctionChanged

    # Not clean if any output is no longer valid
    if not recipe.outputs_valid:
        return Status.OutputsInvalid

    # All checks passed
    return Status.Ok


def brew(recipe: Recipe[R], *, jobs: int) -> R:
    """
    Evaluate a Recipe and all dependent inputs - this will build the computational graph and execute any needed
    dependencies to produce the outputs of the input Recipe

    :param recipe: The Recipe to evaluate
    :param jobs: The number of jobs to use for evaluating the recipe in parallel, 1 job corresponds to no parallelism,
                 zero or negative values will cause alkymi to use the system's default number of jobs
    :return: The outputs of the Recipe (which correspond to the outputs of the bound function)
    """
    graph = create_graph(recipe)
    statuses = compute_recipe_status(recipe, graph)
    result, _ = evaluate_recipe(recipe, graph, statuses, jobs)
    return result
